﻿//     ___                        _                    _     
//    / _ \_   _ _ __  _ __   ___| |_  /\/\   __ _ ___| |_ ___ _ __ 
//   / /_)/ | | | '_ \| '_ \ / _ \ __|/    \ / _` / __| __/ _ \ '__|
//  / ___/| |_| | |_) | |_) |  __/ |_/ /\/\ \ (_| \__ \ ||  __/ |  
//  \/     \__,_| .__/| .__/ \___|\__\/    \/\__,_|___/\__\___|_|  
//              |_|   |_|                    
//                                           
using UnityEngine;

namespace PuppetMaster
{
    public class ChaseCamX : MonoBehaviour
    {
        private ChaseCamX _instance;
        [SkipSerialisation] 
        public ChaseCamX Instance { get { return _instance; } }

        [SkipSerialisation] public GameObject ChaseObj;
        [SkipSerialisation] public PhysicalBehaviour ChasePB;
        [SkipSerialisation] public PhysicalBehaviour ChaseTarget;
        [SkipSerialisation] public PhysicalBehaviour ChaseItem;
        [SkipSerialisation] public Rigidbody2D RealPuppet;

        [SkipSerialisation] public Puppet Puppet;

        [SkipSerialisation] public float yOffset;
        [SkipSerialisation] public float yOffsetTemp = 0f;
        [SkipSerialisation] public float yOffsetPct  = 50f;
        [SkipSerialisation] public float yFactor     = 0f;
        [SkipSerialisation] public float LookAheadX  = 2.0f;
        [SkipSerialisation] public float SpeedAhead  = 2.0f;
        [SkipSerialisation] public float ItemTimeout = 2.0f;
        [SkipSerialisation] public float yThreshold  = 1.0f;
        [SkipSerialisation] public float xThreshold  = 2.0f;
        [SkipSerialisation] public float screenY     = Screen.height;

        [SkipSerialisation] public bool isMoving = false;
        [SkipSerialisation] public bool lockedOn = false;
        private float speedControl = 0.5f;

        private bool setCamera     = false;
        private bool checkedMouse  = false;
        private bool targetClicked = false;
        private bool skipFrame     = false;
        private bool miscBool      = false;
        private int myHash;

        private readonly float maxSpeed = 5.0f;

        private int notMoving2   = 0;
        private int needLock     = 0;
        private int speedFrames  = 0;
        private float notMoving  = 0f;
        private float miscFloat  = 0f;
        private float miscFloat2 = 0f;
        private float avgSpeed   = 0f;
        private float lockedY    = 0f;

        private readonly float slerpSpeed = 5;

        private ChaseModes ChaseMode = ChaseModes.Idle;

        private Vector3 velocity = Vector3.zero;
        private Vector3 cTarget;
        private Vector3 cItem;
        private Vector3 cVehicle;
        private Vector3 cPuppet;
        private Vector3 cItemStill;

        private Camera GCamera;
        private bool CustomCameraInitialized = false;
        private bool skipOrtho               = false;
        private float cameraSize             = 0;

        private string debugID                      = "<color=red>ChaseCam</color>: ";
        public CustomModes CustomMode        = CustomModes.Off;


        public enum ChaseModes
        {
            Idle,
            Puppet,
            Item,
            Action,
            Vehicle,
        }

        public enum CustomModes
        {
            Off,
            GroundPound,
            SonicBoom,
            BikeThrow,
        }


        //
        // ─── UNITY AWAKE ────────────────────────────────────────────────────────────────
        //
        public void Awake()
        {
            if (_instance != null && _instance != this)
            {
                Util.Notify("Destroying 2nd Chasecam", VerboseLevels.Full);
                Destroy(this.gameObject);
            } else
            {
                _instance   = this;
                debugID     = "<color=green>ChaseCam " + GetHashCode().ToString().Substring(1,3) + "</color>: ";
            }
            
            CreateChaseObject();

            GCamera = Global.main.camera;
            myHash  = GetHashCode();

        }


        //
        // ─── UNITY LATE UPDATE ────────────────────────────────────────────────────────────────
        //
        public void LateUpdate()
        {
            if (ChaseMode == ChaseModes.Idle || (ChaseTarget == null && ChaseMode != ChaseModes.Item)) return;

            if (skipFrame)
            {
                skipFrame = false;
                return;
            }

            if (setCamera)
            {
                setCamera = false;
                if (!Global.main.CameraControlBehaviour.CurrentlyFollowing.Contains(ChasePB))
                {
                    Global.main.CameraControlBehaviour.CurrentlyFollowing.Clear();
                    Global.main.CameraControlBehaviour.CurrentlyFollowing.Add(ChasePB);
                }

                return;
            }

            if (CustomMode != CustomModes.Off)
            {
                MoveCustomCamera();
                return;
            }

            if (CustomCameraInitialized)
            {
                Util.Notify(debugID + "Custom Cam Off", VerboseLevels.Full);

                //  Disable custom cam, go back to default

                Global.main.CameraControlBehaviour.enabled = true;

                CustomCameraInitialized = false;

                Time.timeScale = 1f;

            }


            //  Clean up rogue chasecams
            if (Time.frameCount % 1000 == 0)
            {
                if ((PuppetMaster.ChaseCam != null && PuppetMaster.ChaseCam != this) ||
                    !Global.main.CameraControlBehaviour.CurrentlyFollowing.Contains(ChasePB))
                {
                    Util.Notify(debugID + "Destroyed", VerboseLevels.Full);

                    if (ChasePB  != null) UnityEngine.Object.Destroy(ChasePB);
                    if (ChaseObj != null) UnityEngine.Object.Destroy(ChaseObj);

                    this.enabled = false;
                    UnityEngine.Object.Destroy(this);
                }
            }

            //  Check if zoom level has changed
            if (Time.frameCount % 10 == 0 && cameraSize != GCamera.orthographicSize)
            {
                if (Global.main.CameraControlBehaviour.CurrentlyFollowing.Contains(ChasePB)) {
                    Util.Notify(debugID + "Zoom <color=yellow>" + GCamera.orthographicSize + "</color>", VerboseLevels.Full);
                    SetChaseBounds();
                }
            }

            //  Prevent chasecam following puppet when its being dragged around the screen
            if (KB.MouseDown)
            {
                if (!checkedMouse)
                {
                    checkedMouse = true;

                    Collider2D[] NoCollide = Puppet.PBO.transform.root.GetComponentsInChildren<Collider2D>();

                    foreach (Collider2D collider in NoCollide)
                    {
                        if (!collider || collider == null) continue;

                        if ((bool)collider?.OverlapPoint((Vector2)Global.main.MousePosition))
                        {
                            targetClicked = true;
                            break;
                        }
                    }

                    if (Puppet.IsInVehicle)
                    {
                        if (Garage.Bike?.state != PuppetBike.BikeStates.Ready)
                            NoCollide = Garage.Bike.BikeT.GetComponentsInChildren<Collider2D>();
                        else if (Garage.Car?.state == PuppetCar.CarStates.Ready)
                            NoCollide = Garage.Car.CarT.GetComponentsInChildren<Collider2D>();
                        foreach (Collider2D collider in NoCollide)
                        {
                            if (!collider || collider == null) continue;

                            if ((bool)collider?.OverlapPoint((Vector2)Global.main.MousePosition))
                            {
                                targetClicked = true;
                                break;
                            }
                        }
                    }
                }

                if (targetClicked) return;
            }
            else checkedMouse = targetClicked = false;

            //  Chase funcs
            if (ChaseMode == ChaseModes.Puppet) TrackPuppet();
            else if (ChaseMode == ChaseModes.Item) TrackItem();
            else if (ChaseMode == ChaseModes.Vehicle) TrackVehicle();



        }

        //
        // ─── TRACK PUPPET ────────────────────────────────────────────────────────────────
        //
        private void TrackPuppet()
        {
            if (Puppet == null)
            {
                ChaseMode = ChaseModes.Idle;
                return;
            }
            cTarget    = ChaseObj.transform.position;
            cPuppet    = ChaseTarget.transform.position;

            if (Puppet.IsInVehicle && Puppet.ChaseVehicle)
            {
                lockedY     = cPuppet.y + yOffset;
                speedFrames = 0;
                avgSpeed    = 0;
                ChaseMode   = ChaseModes.Vehicle;

                Util.Notify(debugID + "Chasing <color=yellow>Vehicle</color>", VerboseLevels.Full);
                return;
            }

            //  have camera look ahead the direction puppet is facing
            cPuppet.x += LookAheadX * -Puppet.Facing;

            //  have camera Keep the chosen Y offset from the players position
            cPuppet.y += (yOffset + yOffsetTemp);

            float diffX   = Mathf.Abs(cTarget.x - cPuppet.x);
            float diffY   = Mathf.Abs(cTarget.y - cPuppet.y);
            float camtude = GCamera.velocity.magnitude;

            float vDistance = Vector3.SqrMagnitude(cTarget - cPuppet);
            float vMaxSpeed = Mathf.Max(maxSpeed, vDistance);
            float rpvy = RealPuppet.velocity.y;
            
            if (rpvy < -1)  ++notMoving2; 
            else notMoving2 = 0;

            
            if (notMoving2 > 10) cPuppet.y = ChaseTarget.transform.position.y + (cPuppet.y - ChaseTarget.transform.position.y);
            if (RealPuppet.velocity.normalized.magnitude > (ChaseObj.transform.position.normalized.magnitude)  ) { needLock = 0; speedControl = Mathf.Clamp(speedControl + 0.02f, 0.1f, 1f); }
            else if(++needLock > 20) speedControl = Mathf.Clamp(speedControl - 0.02f, 0.1f, 1f);
            velocity *= vDistance * Time.deltaTime;
            //ModAPI.Notify(speedControl);
            //cPuppet.y = cTarget.y;

            if (isMoving || diffX > xThreshold || diffY > yThreshold)
            {

                isMoving = true;
                ChaseObj.transform.position = Vector3.SmoothDamp(
                    Util.ClampPos(cTarget),
                    cPuppet,
                    ref velocity,
                    Time.deltaTime,
                    vMaxSpeed * speedControl
                );

                if (camtude < 1f && ++notMoving > 50) isMoving = false;
            }
        }
        

        //
        // ─── TRACK VEHICLE ────────────────────────────────────────────────────────────────
        //
        private void TrackVehicle()
        {
            if (!Puppet.IsInVehicle || !Puppet.ChaseVehicle)
            {
                Util.Notify(debugID + "Chasing <color=yellow>Puppet</color>", VerboseLevels.Full);

                SetPuppet(Puppet, false);
                ChaseMode = ChaseModes.Puppet;
                return;
            }

            if (SpeedAhead < 1) return;

            if (++speedFrames % 10 == 0) { speedFrames = 1; avgSpeed = 0f; }

            avgSpeed += SpeedAhead;


            cTarget  = ChaseObj.transform.position;
            cVehicle = ChaseTarget.transform.position;

            //  have camera look ahead the direction puppet is facing
            cVehicle.x += (avgSpeed / speedFrames) * -Puppet.Facing;

            //  have camera Keep the chosen Y offset from the players position
            cVehicle.y += yOffset;

            if (Mathf.Abs(cVehicle.y - lockedY) < 5f) cVehicle.y = lockedY;

            ChaseObj.transform.position = cVehicle;
        }



        //
        // ─── TRACK ITEM ────────────────────────────────────────────────────────────────
        //
        private void TrackItem()
        {
            if (ChaseTarget == null)
            {
                if (notMoving == 0) notMoving = Time.time;
                if (Time.time - notMoving > 2f) StopQuickChase();
                return;
            }

            if (KB.AnyKey)
            {
                StopQuickChase();
                return;
            }

            cTarget = ChaseObj.transform.position;
            cItem   = Util.ClampPos(ChaseTarget.transform.position);

            float vDistance = Vector3.SqrMagnitude(cItemStill - cItem);

            if (vDistance > 0.01f)
            {
                notMoving  = Time.time;
                cItemStill = cTarget;
            }
            else if (Time.time - notMoving > 2f)
            {
                StopQuickChase();
                return;
            }

            ChaseObj.transform.position = cItem;

        }


        //
        // ─── QUICK CHASE ────────────────────────────────────────────────────────────────
        //
        public void QuickChase(PhysicalBehaviour target)
        {
            Util.Notify(debugID + "Chasing <color=yellow>Item</color>", VerboseLevels.Full);
            ChaseTarget = target;
            ChaseMode   = ChaseModes.Item;
            cItemStill  = Vector3.zero;
        }


        public void StopQuickChase()
        {
            if (KB.Modifier) return;
            if (PuppetMaster.isMaxPayne) Util.MaxPayne(false);

            notMoving = 0;
            lockedOn  = false;

            if (Puppet != null)
            {
                Util.Notify(debugID + "Chasing <color=yellow>Puppet</color>", VerboseLevels.Full);

                skipOrtho = true;
                ChaseMode = ChaseModes.Puppet;
                SetPuppet(Puppet, false);
                return;
            }

            ChaseMode = ChaseModes.Idle;
        }

        //
        // ─── MOVE CUSTOM CAMERA ────────────────────────────────────────────────────────────────
        //
        private void MoveCustomCamera()
        {
            if (!CustomCameraInitialized)
            {
                if (!Global.main.CameraControlBehaviour.CurrentlyFollowing.Contains(ChasePB))
                {
                    CustomMode = CustomModes.Off;
                    ChaseMode  = ChaseModes.Idle;
                    return;
                }

                GCamera = Global.main.camera;
                Global.main.CameraControlBehaviour.enabled = false;
                CustomCameraInitialized = true;
                miscBool = false;

                miscFloat  = Time.time + 10f;
                miscFloat2 = Time.time + 2f;

            }

            if (CustomMode == CustomModes.BikeThrow)
            {
                Time.timeScale = 0.5f;
                cTarget = ChaseObj.transform.position;
                cPuppet = ChaseTarget.transform.position;

                ChaseObj.transform.position = Vector3.Slerp(cTarget, cPuppet, slerpSpeed * Time.deltaTime);
                Global.CameraPosition = Util.ClampPos(cTarget);

                miscFloat = Mathf.Abs(ChaseItem.transform.position.x - ChaseTarget.transform.position.x);

                float btzoomOut = Mathf.Clamp(miscFloat, GCamera.orthographicSize, 40);
                GCamera.orthographicSize = Mathf.Lerp(GCamera.orthographicSize, btzoomOut, Time.deltaTime * 2);

                if (Time.time > miscFloat2)
                {
                    CustomMode = CustomModes.Off;
                }

                return;
            }

            cTarget = ChaseObj.transform.position;
            cPuppet = ChaseTarget.transform.position;

            ChaseObj.transform.position = Vector3.Slerp(cTarget, cPuppet, slerpSpeed * Time.deltaTime);
            Global.CameraPosition = Util.ClampPos(cTarget);

            float zoomOut = Mathf.Clamp(Puppet.Actions.jump.sonicForce * 3, GCamera.orthographicSize, 40);
            GCamera.orthographicSize = Mathf.Lerp(GCamera.orthographicSize, zoomOut, Time.deltaTime * 2);

            if (CustomMode == CustomModes.SonicBoom)
            {
                Time.timeScale = 0.3f;
                if (!miscBool)
                {
                    miscBool  = true;
                    miscFloat = Time.time + 2f;
                }

                if (KB.AnyKey && !KB.Down) CustomMode = CustomModes.Off;
            }

            if (Time.time > miscFloat)
            {
                CustomMode = CustomModes.Off;
            }

        }

        //
        // ─── SET PUPPET ────────────────────────────────────────────────────────────────
        //
        public void SetPuppet(Puppet puppet, bool resetY = false)
        {
            if (puppet == null) return;
            
            CustomMode = CustomModes.Off;
            Puppet     = PuppetMaster.Puppet;

            if (ChaseObj == null) CreateChaseObject();

            ChaseTarget = Puppet.LB?["MiddleBody"].PhysicalBehaviour;

            //  Set the constant Yoffset
            if (resetY)
            {
                cameraSize = GCamera.orthographicSize;

                yOffset = Global.main.CameraControlBehaviour.transform.position.y - ChaseTarget.transform.position.y;

                yFactor = yOffset / GCamera.orthographicSize;

                Util.Notify(debugID + "yOffset: <color=yellow>" + yOffset + "</color> yFactor: <color=yellow>" + yFactor + "</color>", VerboseLevels.Full);
            }


            //  Start invisible tracker at puppet position
            Vector3 cPuppet = ChaseTarget.transform.position;

            //  have camera look ahead the direction puppet is facing
            cPuppet.x += LookAheadX * -Puppet.Facing;

            //  have camera Keep the chosen Y offset from the players position
            cPuppet.y += yOffset;

            ChaseObj.transform.position = cPuppet;

            ChaseMode = ChaseModes.Puppet;
            setCamera = true;
            lockedOn  = false;
            skipFrame = true;

            if (!skipOrtho) SetChaseBounds();
            else skipOrtho = false;

            RealPuppet = ChaseTarget.rigidbody;
        }


        public void SetChaseBounds()
        {
            if (!Global.main.CameraControlBehaviour.CurrentlyFollowing.Contains(ChasePB)) { return; }

            cameraSize = GCamera.orthographicSize;
            yOffset = cameraSize * yFactor;

            //  Start invisible tracker at puppet position
            Vector3 cPuppet = ChaseTarget.transform.position;

            //  have camera look ahead the direction puppet is facing
            cPuppet.x += LookAheadX * -Puppet.Facing;

            //  have camera Keep the chosen Y offset from the players position
            cPuppet.y += yOffset;

            ChaseObj.transform.position = cPuppet;

        }

        //
        // ─── CREATE CHASE OBJECT ────────────────────────────────────────────────────────────────
        //
        public void CreateChaseObject()
        {
            //  Destroy any rogue chase objects
            foreach (PhysicalBehaviour PB in Global.main.PhysicalObjectsInWorld)
            {
                if (PB.name == "ChaseCam")
                {
                    Util.Notify(debugID + "Destroying Obj <color=yellow>" + PB.GetHashCode() + "</color>", VerboseLevels.Full);

                    Destroy(PB);
                }
            }

            //  Create a hidden chase-cam object that the camera follows
            Texture2D tex   = new Texture2D(0, 0);
            Sprite mySprite = Sprite.Create(tex, new Rect(0.0f, 0.0f, tex.width, tex.height), new Vector2(0.1f, 0.1f), 0.1f);
            ChaseObj        = ModAPI.CreatePhysicalObject("ChaseCam", mySprite);

            // Initialize chase-cam object
            ChaseObj.transform.position = Global.CameraPosition;

            ChasePB = ChaseObj.GetOrAddComponent<PhysicalBehaviour>();

            ChasePB.MakeWeightless();
            ChasePB.Selectable = false;
            ChasePB.Deletable  = true;

            //  Prevent all collisions with invisible chase object
            ChaseObj.SetLayer(10);

            //  Prevent this from feking up anyones contraption
            ChaseObj.GetOrAddComponent<Optout>();

            PuppetMaster.ChaseCam = this;

            Util.Notify(debugID + "Created Chase Obj: <color=yellow>" + ChasePB.GetHashCode() + "</color>" , VerboseLevels.Full);
        }

        //
        // ─── STOP (self-destruct) ────────────────────────────────────────────────────────────────
        //
        public void Stop()
        {
            Util.Notify(debugID + "Force Stop", VerboseLevels.Full);

            Global.main.CameraControlBehaviour.CurrentlyFollowing.Clear();

            if (ChasePB != null)  UnityEngine.Object.Destroy(ChasePB);
            if (ChaseObj != null) UnityEngine.Object.Destroy(ChaseObj);

            //this.enabled = false;
            UnityEngine.Object.Destroy(this);
        }

        void OnDestroy()
        {
            if (this == _instance)
            {
                _instance = null;
                if (Global.main.CameraControlBehaviour.CurrentlyFollowing.Contains(ChasePB))
                {
                    Global.main.CameraControlBehaviour.CurrentlyFollowing.Remove(ChasePB);
                }
                if (ChasePB  != null) UnityEngine.Object.Destroy(ChasePB);
                if (ChaseObj != null) UnityEngine.Object.Destroy(ChaseObj);
            }
        }
    }
}