﻿//     ___                        _                    _     
//    / _ \_   _ _ __  _ __   ___| |_  /\/\   __ _ ___| |_ ___ _ __ 
//   / /_)/ | | | '_ \| '_ \ / _ \ __|/    \ / _` / __| __/ _ \ '__|
//  / ___/| |_| | |_) | |_) |  __/ |_/ /\/\ \ (_| \__ \ ||  __/ |  
//  \/     \__,_| .__/| .__/ \___|\__\/    \/\__,_|___/\__\___|_|  
//              |_|   |_|                
//                                           
using System;
using System.Collections.Generic;
using System.Linq;
using UnityEngine;
using UnityEngine.Events;

namespace PuppetMaster
{
    [SkipSerialisation]
    public class Puppet
    {
        public bool IsActive                 = false;
        public bool IsWalking                = false;
        public bool IsCrouching              = false;
        public bool IsFiring                 = false;
        public bool IsHero                   = false;
        public bool IsHeroAiming             = false;
        public bool JumpLocked               = false;
        public bool IsEmote                  = false;
        public bool FacingLeft               = false;
        public bool BlockMoves               = false;
        public bool checkFlipColliders       = false;
        public bool pauseAiming              = false;
        public bool resetFlipControl         = false;
        public bool RunUpdate                = false;
        public bool CanHeroAim               = false;
        public bool HeroMode                 = false;
        public bool hasFlipped               = false;
        public bool ChaseVehicle             = true;
        
        public Hero hero;

        private Liquid LifeLiquid   = Liquid.GetLiquid("LIFE SERUM");

        public bool IsInvincible
        {
            get { return _isInvincible; }
            set { 
                _isInvincible = value; 
                RunUpdate     = value; 
                if (value && !LB["Head"].IsAndroid) RunLimbs(LimbStrenthen);
                else if (!value && !LB["Head"].IsAndroid) {
                    RunLimbs(LimbReset);
                    RunRigids(RigidReset);
                }
            }
        }

        private bool _isInvincible = false;

        public bool IsAiming
        {
            get { return _IsAiming; }
            set { 
                _IsAiming = value;
                if (value == false) CheckMouseClickStatus();
            }
        }

        private bool _IsAiming   = false;
        private bool _IsPointing = false;

        public bool IsPointing
        {
            get { return _IsPointing; }
            set { 
                _IsPointing = value; 
                if (value == false ) CheckMouseClickStatus();
            }
        }


        

        public bool IsInVehicle
        {
            get { return _IsInVehicle; }
            set { 
                _IsInVehicle = value; 
                Inventory.AllowedTypes  = _IsInVehicle 
                                        ? Inventory.ItemTypes.VehicleSafe 
                                        : Inventory.ItemTypes.All;
                }
        }
        private bool _IsInVehicle = false;


        public int EmoteId                   = -1;
        
        public float Facing                  = 1f;
        public float TotalWeight             = 0;
        public float DisableMoves            = 0;
        
        public PersonBehaviour PBO           = null;
        public Inventory Inventory           = null;
        public Actions Actions               = null;
        public Collider2D[] Colliders        = null;

        public Thing HoldingF                = null;
        public Thing HoldingB                = null;

        public GripBehaviour GripF           = null;
        public GripBehaviour GripB           = null;

        public KeyCode IgnoreKey             = KeyCode.None;
        public bool FireProof                = false;
        
        public AudioSource AudioSFX1;

        //

        public Dictionary<string, int> PuppetPose     = new Dictionary<string, int>();
        public Dictionary<string, Rigidbody2D> RB2    = new Dictionary<string, Rigidbody2D>();
        public Dictionary<string, LimbBehaviour> LB   = new Dictionary<string, LimbBehaviour>();
        private readonly Dictionary<string, LimbSnapshot> LimbOriginals   = new Dictionary<string, LimbSnapshot>();
        private readonly Dictionary<string, RigidSnapshot> RigidOriginals = new Dictionary<string, RigidSnapshot>();
        
        public List<Collider2D> FlipColliders = new List<Collider2D>();
        public CustomParts customParts        = new CustomParts();
        public FlipStates FlipState           = FlipStates.ready;

        public class RigidSnapshot
        {
            public float inertia;
            public float mass;
            public float drag;
        }

        



        public bool IsReady    => (bool)(PuppetMaster.Master.Activated 
                                                   && PBO.isActiveAndEnabled 
                                                   && PBO.IsAlive() 
                                                   && PBO.Consciousness >= 1 
                                                   && PBO.ShockLevel < 0.3f );

        public bool IsUpright  => (bool)(!LB["LowerBody"].IsOnFloor
                                                   && !LB["UpperBody"].IsOnFloor
                                                   && !LB["Head"].IsOnFloor
                                                   && LB["FootFront"].IsOnFloor
                                                   && LB["Foot"].IsOnFloor)
                                                   && !IsInVehicle;

        public bool CanWalk    => (bool)(PBO.IsTouchingFloor 
                                                   && !IsWalking 
                                                   && !IsCrouching
                                                   && !KB.Up 
                                                   && !IsInVehicle //&& Actions.attack.state == AttackState.ready
                                                   && Actions.prone.state == ProneState.ready);
        public bool CanAttack       => (bool)(!IsInVehicle && Actions.jump.state == JumpState.ready);
        public bool IsFlipped       => (bool)(PBO.transform.localScale.x < 0.0f);
        public bool DisabledMoves   => (bool)(IsInVehicle || BlockMoves || DisableMoves > Time.time);
        public bool SpecialMode     => (bool)(IsAiming || IsPointing || Actions.CombatMode );


        //
        // ─── INITIALIZE PUPPET ────────────────────────────────────────────────────────────────
        //
        public void Init(PersonBehaviour _pbo)
        {
            TotalWeight       = 0f;
            PBO               = _pbo;

            //  Cache Rigidbody Maps
            Rigidbody2D[] RBs = PBO.transform.GetComponentsInChildren<Rigidbody2D>();
            
            foreach (Rigidbody2D rb in RBs)
            {

                //  Take snapshots of the current values for drag, intertia & mass
                //  Since these are modified during actions, we can set it back proper
                //
                RB2.Add(rb.name, rb);

                RigidSnapshot RBOG = new RigidSnapshot()
                {
                    drag    = rb.drag,
                    inertia = rb.inertia,
                    mass    = rb.mass,
                };

                RigidOriginals.Add(rb.name, RBOG);

                TotalWeight += rb.mass;

                GripBehaviour GB;

                //  @TODO: This shit is broken
                switch (rb.name)
                {
                    case "LowerArmFront":
                        if (rb.TryGetComponent<GripBehaviour>(out GB)) GripF = GB;
                        if (GripF.CurrentlyHolding != null) HoldingF = GripF.CurrentlyHolding.gameObject.GetOrAddComponent<Thing>();
                        break;

                    case "LowerArm":
                        if (rb.TryGetComponent<GripBehaviour>(out GB)) GripB = GB;
                        if (GripB.CurrentlyHolding != null) HoldingB = GripB.CurrentlyHolding.gameObject.GetOrAddComponent<Thing>();
                        break;
                }
            }

            SetupLimbs(true);

            if (PBO.OverridePoseIndex != (int)PoseState.Rest || PBO.OverridePoseIndex != (int)PoseState.Sitting)
            {
                PBO.OverridePoseIndex = -1;
            }

            Actions        = new Actions(this);
            Inventory      = new Inventory(this);
            FacingLeft     = IsFlipped;
            Facing         = FacingLeft ? 1f : -1f;

            Colliders      = PBO.transform.root.GetComponentsInChildren<Collider2D>(); 

            Util.Notify("Puppet set: <color=yellow>" + PBO.name + "</color>", VerboseLevels.Minimal);

            PuppetMaster.Puppet = this;
            PuppetMaster.StartChaseCam();


            Garage.Puppet  = this;

            AudioSFX1 = PBO.gameObject.AddComponent<AudioSource>();

            //  Find any custom attached armor/clothes/shtuff connected to puppet
            CheckAttachments();

            if (Hero.HeroCheck()) {
                Util.Notify("<color=red>A <color=orange>hero<color=red> was found!</color>", VerboseLevels.Minimal);
                IsHero = true;
                if (hero == null) hero = new Hero();
                hero.LoadHero();
            }
        }

        public void Reset()
        {
            CheckAttachments();
            SetupLimbs();
        }

        public void SetupLimbs(bool init=false)
        {
            //  Cache LimbBeheaviour Maps
            LimbBehaviour[] LBs = PBO.GetComponentsInChildren<LimbBehaviour>();
            
            LB.Clear();
            LimbOriginals.Clear();

            foreach (LimbBehaviour limb in LBs)
            {
                LB.Add(limb.name, limb);

                LimbSnapshot LBOG = new LimbSnapshot()
                {
                    BaseStrength      = limb.BaseStrength,
                    BreakingThreshold = limb.BreakingThreshold,
                    Broken            = limb.Broken,
                    BruiseCount       = limb.BruiseCount,
                    Frozen            = limb.Frozen,
                    Health            = limb.Health,
                    Numbness          = limb.Numbness,
                    RegenerationSpeed = limb.RegenerationSpeed,
                    Vitality          = limb.Vitality,
                    Properties        = limb.GetComponent<PhysicalBehaviour>().Properties.ShallowClone(),
                };

                if (limb.HasJoint)
                {
                    LBOG.jLimitMax = limb.Joint.limits.max;
                    LBOG.jLimitMin = limb.Joint.limits.min;
                }

                LimbOriginals.Add(limb.name, LBOG);

                if (init)
                {
                    Hero.SetupLimbButton(limb);
                }
            }
        }
        

        //
        // ─── UNITY FIXED UPDATE ────────────────────────────────────────────────────────────────
        //
        public void FixedUpdate()
        {
            //if (!IsInvincible) IsInvincible = true;

            if (Time.frameCount % 100 == 0) 
            {
                if (PBO == null) 
                { 
                    this.IsActive = false;
                    PuppetMaster.ActivePuppet = 0;
                    return;
                }

                if (resetFlipControl && !KB.Left && !KB.Right && (bool)!HoldingF?.isLodged) {
                    
                    resetFlipControl = false;

                }
            }

            Actions.RunActions();

            if (FlipState != FlipStates.ready) RunFlip();

            if (IsAiming) {
                if (IsReady) AimWeapon();
            } else if (IsPointing)
            {
                if (IsReady) PointWeapon();
            }

            if (IsWalking)
            {
                if (KB.Left && !FacingLeft && (Time.time - KB.KeyTimes.Left > 0.5f) && (Time.time - KB.KeyTimes.Left < 1.0f) && Flip())        FacingLeft = true;
                else if (KB.Right && FacingLeft && (Time.time - KB.KeyTimes.Right > 0.5f) && (Time.time - KB.KeyTimes.Right < 1.0f) && Flip())  FacingLeft = false;
                else
                {
                    IsWalking = false;
                    StopPerson();
                }
            }

            if (Inventory.TriggerPickup)
            {
                Inventory.TriggerPickup = false;

                HoldThing(Inventory.Clone);

                PlayInventorySound();
            }

            if (KB.Inventory) Inventory.DoInventory();

            if (Inventory.InventoryTriggered && !KB.Inventory) Inventory.RunInventory();

            if (checkFlipColliders && Time.frameCount % 5 == 0) CheckClearedCollisions();
        }




        //
        // ─── UNITY UPDATE ────────────────────────────────────────────────────────────────
        //
        public void Update()
        {
            RunUpdate = false;
            if (FireProof)
            {
                RunUpdate           = true;
                PBO.PainLevel       = 0.0f;
                PBO.ShockLevel      = 0.0f;
                PBO.AdrenalineLevel = 1.0f;
            }

            if (IsInvincible)
            {
                RunUpdate           = true;
                PBO.PainLevel       = 0.0f;
                PBO.ShockLevel      = 0.0f;
                PBO.AdrenalineLevel = 1.0f;

                RunLimbs(LimbGangster);
            }
        }

        public void LateUpdate()
        {
            if (IsHero && HeroMode) hero.HeroAimRun();
        }

        public void RefreshInventory(Inventory inventory)
        {
            Inventory = inventory;
        }


        public void CheckAttachments()
        {
            foreach (FixedJoint2D fJoint in UnityEngine.Object.FindObjectsOfType<FixedJoint2D>())
            {
                if (RB2.Values.Contains(fJoint.connectedBody))
                {
                    if (customParts.Add(fJoint.gameObject, fJoint.connectedBody))
                    {
                        fJoint.attachedRigidbody.mass *= 0.01f;
                        Util.Notify("Found attached: " + fJoint.gameObject.name, VerboseLevels.Full);
                    }
                }
            }
        }

        public void SetHeroAttack(int attackId, string limbName)
        {
            if (hero == null) hero = new Hero();

            hero.HeroSetup(attackId, limbName);
        }

        //
        // ─── FLIP ────────────────────────────────────────────────────────────────
        //
        public void RunFlip()
        {
            if (FlipState == FlipStates.preflip)
            {
                FlipState = FlipStates.flip;

                customParts.PreFlip();
                
                return;
            }

            if (FlipState == FlipStates.flip)
            {
                FlipState = FlipStates.postflip;

                const string limbNamesList = "LowerBody,MiddleBody,UpperBody,UpperArm,UpperArmFront,Head";

                Vector3 flipScale = PBO.transform.localScale;
                Vector3 moveA;
                Vector3 moveB;
                Vector2 moveDif;

                flipScale.x *= -1;

                moveB = RB2["Head"].transform.position;

                foreach (LimbBehaviour limb in LB.Values)
                {
                    if (limb.HasJoint)
                    {
                        limb.BreakingThreshold *= 8;

                        if (!limbNamesList.Contains(limb.name))
                        {
                            JointAngleLimits2D t = limb.Joint.limits;
                            t.min *= -1f;
                            t.max *= -1f;
                            
                            limb.Joint.limits        = t;
                            limb.OriginalJointLimits = new Vector2(limb.OriginalJointLimits.x * -1f, limb.OriginalJointLimits.y * -1f);
                        }
                    }
                }

                PBO.transform.localScale = flipScale;

                moveA   = RB2["Head"].transform.position;
                moveDif = moveB - moveA;

                PBO.AngleOffset *= -1f;
                PBO.transform.position = new Vector2(PBO.transform.position.x + moveDif.x, PBO.transform.position.y);

                foreach (LimbBehaviour limb in PBO.Limbs) if (limb.HasJoint) limb.Broken = false;

                if (HoldingF != null)
                {
                    if (LB["LowerArm"].GripBehaviour.isHolding)
                    {
                        LB["LowerArm"].SendMessage("Use", (object)new ActivationPropagation(), SendMessageOptions.DontRequireReceiver);
                    }
                    LB["LowerArmFront"].SendMessage("Use", (object)new ActivationPropagation(), SendMessageOptions.DontRequireReceiver);
                    HoldingF.ResetPosition();
                }

                FacingLeft = IsFlipped;
                Facing     = FacingLeft ? 1f : -1f;

                customParts.Flip();

                return;
            }

            if (FlipState == FlipStates.postflip)
            {
                FlipState = FlipStates.ready;

                customParts.PostFlip();
                CheckFlipCollision();

                return;
            }
        }

        public bool Flip(bool forced=false)
        {
            if (!forced) { 
                if (
                    DisabledMoves || 
                    !IsUpright || 
                    (KB.Left && KB.Right) || 
                    Actions.jump.state != JumpState.ready || 
                    Actions.backflip.state != BackflipState.ready ||
                    !PBO.isActiveAndEnabled || 
                    !PBO.IsAlive() || 
                    !PBO.IsTouchingFloor || 
                    PBO.Consciousness < 1) return false;

                if ((HoldingF != null && HoldingF.isLodged) || resetFlipControl) {
                    resetFlipControl = true;
                    return false;

                }
            }

            if (FlipState != FlipStates.ready) return false;
            FlipState = FlipStates.preflip;

            return true;
        }


        public void PlayInventorySound()
        {
            if (HoldingF.InvSFX != null) { 
                AudioSFX1.volume = 0.5f;
                AudioSFX1.PlayOneShot(HoldingF.InvSFX);
            }
        }

        public void CheckFlipCollision()
        {
            Collider2D[] noCollide = PBO.transform.root.GetComponentsInChildren<Collider2D>();
            bool isClear;
            Collider2D[] tmpCol;

            if (HoldingF != null)
            {
                Collider2D[] itemcollide = HoldingF.P.GetComponents<Collider2D>();

                foreach (Collider2D coll in itemcollide)
                {
                    coll.enabled = false;
                    isClear = true;

                    tmpCol = Physics2D.OverlapBoxAll(coll.transform.position, (Vector2)coll.bounds.size, coll.transform.eulerAngles.y);
                    if (tmpCol.Length > 0)
                    {
                        foreach (Collider2D col in tmpCol)
                        {
                            PersonBehaviour hitperson = col.attachedRigidbody.GetComponentInParent<PersonBehaviour>();
                            if (hitperson != null && hitperson != PBO)
                            {
                                checkFlipColliders = true;
                                FlipColliders.Add(col);
                                col.enabled = false;
                               // isClear = false;

                            }
                        }
                    }
                    coll.enabled = isClear;
                }

            }

            foreach (Collider2D coll in noCollide) {
                coll.enabled = false;
                isClear      = true;

                tmpCol = Physics2D.OverlapBoxAll(coll.transform.position, (Vector2)coll.bounds.size, 0 );
                if (tmpCol.Length > 0) 
                {
                    foreach (Collider2D col in tmpCol) 
                    { 
                        PersonBehaviour hitperson = col.attachedRigidbody.GetComponentInParent<PersonBehaviour>();
                        if (hitperson != null && hitperson != PBO)
                        {
                            checkFlipColliders = true;
                            FlipColliders.Add(coll);
                            isClear = false;
                        }
                    }
                }
                coll.enabled = isClear;
            }

            return;
        }

        public bool CheckClearedCollisions()
        {
            Collider2D tmpCol;

            for (int i = FlipColliders.Count; --i >= 0;)
            {
                Collider2D coll = (Collider2D)FlipColliders[i];

                if (coll == null)
                {
                    FlipColliders.RemoveAt(i);
                    continue;
                }
                if (tmpCol = Physics2D.OverlapBox(coll.transform.position, (Vector2)coll.bounds.size, 0))
                {
                    PersonBehaviour hitperson = tmpCol.attachedRigidbody.GetComponentInParent<PersonBehaviour>();
                    if (hitperson == null || hitperson == PBO)
                    {
                        coll.enabled = true;
                        FlipColliders.RemoveAt(i);
                    }
                }
            }

            checkFlipColliders = FlipColliders.Count > 0;
            return checkFlipColliders;
        }

        //
        // ─── CHECK CONNTROLS ────────────────────────────────────────────────────────────────
        //
        public void CheckControls()
        {
            if (!KB.ActionHeld && !KB.MouseDown) IsFiring = false;


            if (KB.Activate && HoldingF != null)
            {
                //  ---
                //  Activate Items
                //  ---
                HoldingF.Activate(HoldingF.isAutomatic);
            }
            if(KB.Action2 || KB.Action || ( (Actions.CombatMode || IsPointing) && (KB.MouseDown || KB.Mouse2Down)))
            {
                //  ---
                //  Action Keys
                //  ---
                if (HoldingF != null) { 
                    if (HoldingF.isChainSaw)
                    {
                        HoldingF.Activate(true);
                    }
                    if (HoldingF.canStrike ) {

                        if (HoldingF.isLodged)
                        {
                            // Sword or knife is stuck in enemy
                            if (KB.Left || KB.Right)
                            {
                                if (KB.Left == FacingLeft && KB.Right != FacingLeft) Actions.attack.Init(Attack.AttackIds.dislodgeKick);
                                else Actions.attack.Init(Attack.AttackIds.dislodgeBack);
                            } 
                            else if (KB.Down)
                            {
                                if (HoldingF.canSlice) Actions.attack.Init(Attack.AttackIds.dislodgeIchi);
                                else Actions.attack.Init(Attack.AttackIds.dislodgeKick);
                            }
                        } else
                        {
                            if (KB.Action  || (Actions.CombatMode && KB.MouseDown)) Actions.attack.Init(Attack.AttackIds.club);
                            if (KB.Action2 || (Actions.CombatMode && KB.Mouse2Down)) Actions.attack.Init(Attack.AttackIds.thrust);
                        }
                    }
                    
                }
            }
            if (( KB.Left || KB.Right) && !DisabledMoves && IsReady)
            {
                //  ---
                //  Keys LEFT + RIGHT
                //  ---
                if ( CanWalk )
                {

                    if ((KB.KeyCombos.DoubleRight || KB.KeyCombos.DoubleLeft))
                    {
                        if (FacingLeft == KB.Left)
                        {
                            Actions.dive.Init();
                        } else
                        {
                            Actions.backflip.Init();
                        }

                        KB.KeyCombos.DoubleRight = KB.KeyCombos.DoubleLeft = false;
                        
                        
                    } else { 
                        if (!IsWalking)
                        {

                            if (FacingLeft != KB.Left) {
                                
                                if (KB.Left  && (Time.time - KB.KeyTimes.Left > 0.5f) && (Time.time - KB.KeyTimes.Left < 1.0f))   IsWalking   = Flip();
                                if (KB.Right && (Time.time - KB.KeyTimes.Right > 0.5f) && (Time.time - KB.KeyTimes.Right < 1.0f)) IsWalking   = Flip();

                            }
                            else IsWalking  = true;
                            
                            if (IsWalking) PBO.OverridePoseIndex = (int)PoseState.Walking;

                        }

                    }
                } 
                else  if (IsCrouching)
                {
                    if (FacingLeft != KB.Left) Flip();
                    //FacingLeft = IsFlipped;
                }
            } 

            else if (KB.Down && !IsCrouching && Actions.prone.state == ProneState.ready && PBO.IsTouchingFloor && IsReady)
            {
                //  ---
                //  Keys DOWN
                //  ---
                if (KB.KeyCombos.DoubleDown)
                {
                    if (!DisabledMoves) Actions.prone.Init();
                    KB.KeyCombos.DoubleDown = false;
                } 
                else if (!DisabledMoves && Actions.prone.state == ProneState.ready) { 

                    if (KB.Modifier) PBO.OverridePoseIndex = (int)PoseState.Sitting;
                    else
                    {
                        //PBO.OverridePoseIndex  = PuppetPose["JT_Duck"];
                        if (!IsCrouching) Actions.crouch.Init();
                    }

                }

            }

            else if (KB.Up)
            {
                //  ---
                //  Keys UP
                //  ---
                if (!IsInVehicle && Actions.prone.state != ProneState.ready)
                {
                    StopPerson();
                    JumpLocked = true;
                } 
                else if (!IsReady)
                {
                    Recover();
                }
                else if (!DisabledMoves && !IsInVehicle) Actions.jump.Init();
                
            }

            
            else if (KB.Emote)
            {
            }



            else if (KB.Throw && IsReady)
            {
                //  ---
                //  Keys Throw
                //  ---
                if (IsReady && Actions.throwItem.state == ThrowState.ready) Actions.throwItem.Init();

            }


            else {
                
                if (JumpLocked) JumpLocked = false;

            }

            if (KB.Aim)
            {
                //  ---
                //  Keys AIM
                //  ---
                if (KB.Modifier && IsHero)
                {
                    HeroMode = !HeroMode;

                    Util.Notify(" <color=blue>Hero Mode:<color=yellow> " + (HeroMode ? "On":"Off") + "</color>", VerboseLevels.Minimal);

                    if (HeroMode) hero.HeroAimStart();
                    else          hero.HeroAimStop();
                        
                    return;
                }

                if (HeroMode) { HeroMode = false; hero.HeroAimStop(); }

                if (!IsReady) IsAiming = false;

                IsAiming = !IsAiming;

                if (IsAiming) AimingInit();

                else AimingStop();

                if (HoldingF != null && !HoldingF.canShoot)
                {
                    if (HoldingF.isChainSaw || (IsInVehicle && HoldingF.canStab))
                    {
                        Actions.CombatMode = false;
                        IsPointing         = !IsPointing;
                        if (IsPointing && !IsInVehicle) KB.DisableMouse();
                        else {
                            CheckMouseClickStatus();
                            HoldingF.ChangeJoint(Thing.JointTypes.FixedJoint);
                        }

                        Util.Notify("<color=red>POINT MODE: </color>" + (IsPointing ? "ON" : "OFF"), VerboseLevels.Full);

                    } else
                    {
                        if (HoldingF.JointType == Thing.JointTypes.HingeJoint) HoldingF.ChangeJoint(Thing.JointTypes.FixedJoint);
                        
                        IsPointing         = false;
                        Actions.CombatMode = !Actions.CombatMode;
                        
                        Util.Notify("<color=red>COMBAT MODE: </color>" + (Actions.CombatMode ? "ON" : "OFF"), VerboseLevels.Minimal);
                    }
                } else
                {
                    if (HoldingF == null)
                    {
                        IsAiming = IsPointing = Actions.CombatMode = false;
                    }
                }

            }

            if (KB.ActionHeld || (IsAiming && KB.MouseDown))
            {
                //  ---
                //  Keys ACTION
                //  ---
                if (!IsFiring && IsReady)
                {
                    // Check if held item is something aimable
                    if (!LB["LowerArmFront"].isActiveAndEnabled || !LB["LowerArmFront"].GripBehaviour.isHolding || !HoldingF.canShoot)
                    {
                        IsFiring = false;
                        return;
                    }

                    IsFiring = true;

                    FireWeapon();
                }
                else if (HoldingF.isAutomatic) FireWeapon();
            }

            //if (KB.Alt && HoldingF != null) HoldingF.PrintPosition();
            
        }


        //
        // ─── DO EMOTE ────────────────────────────────────────────────────────────────
        //
        public void DoEmote()
        {
            if (!IsEmote)
            {
                EmoteId = -1;

                KB.DisableNumberKeys();

                IsEmote = true;
            }

            EmoteId = KB.CheckNumberKey();

            if (EmoteId == -1) return;

        }

        
        //
        // ─── FIRE WEAPON ────────────────────────────────────────────────────────────────
        //
        private void FireWeapon()
        {
            if (!PBO.IsAlive()) return;
            HoldingF.Activate(HoldingF.isAutomatic);
        }


        //
        // ─── STOP PERSON ────────────────────────────────────────────────────────────────
        //
        public void StopPerson()
        {
            if (!IsWalking && !IsCrouching) { 

                if (PBO.OverridePoseIndex != (int)PoseState.Rest || PBO.OverridePoseIndex != (int)PoseState.Sitting)
                {
                    PBO.OverridePoseIndex = -1;
                }
            }
        }


        //
        // ─── AIMING INIT ────────────────────────────────────────────────────────────────
        //
        public void AimingInit()
        {
            // Check if held item is something that puppet can point
            if (!LB["LowerArmFront"].isActiveAndEnabled || !LB["LowerArmFront"].GripBehaviour.isHolding || HoldingF == null || !HoldingF.canShoot)
            {
                //IsAiming = false;
                AimingStop();
                CheckMouseClickStatus();

                return;
            }

            IsAiming = HoldingF.canShoot;
            
            if (IsAiming) KB.DisableMouse();


            

            if (HoldingF.holdToSide)
            {
                //HoldingF.SetHoldingPosition();

                LB["LowerArmFront"].SendMessage("Use", (object)new ActivationPropagation(), SendMessageOptions.DontRequireReceiver);
                
                HoldingF.ResetPosition(true);
            }


        }

        //
        // ─── AIMING STOP ────────────────────────────────────────────────────────────────
        //
        public void AimingStop()
        {
            IsAiming = false;
            
            LB["LowerArmFront"].Broken =
            LB["UpperArmFront"].Broken = false;

            CheckMouseClickStatus();

            if (HoldingF != null && HoldingF.holdToSide) {

                LB["LowerArmFront"].SendMessage("Use", (object)new ActivationPropagation(), SendMessageOptions.DontRequireReceiver);
                HoldingF.ResetPosition();

            }

        }


        //
        // ─── AIM WEAPON ────────────────────────────────────────────────────────────────
        //
        public void AimWeapon()
        {
            //
            //  Reference: https://gamedevbeginner.com/make-an-object-follow-the-mouse-in-unity-in-2d/
            //
            Vector2 mouse = Camera.main.ScreenToWorldPoint(Input.mousePosition);


            if (mouse.x * Facing > (RB2["Head"].position.x * Facing))
            {
                LB["LowerArmFront"].Broken =
                LB["UpperArmFront"].Broken = false;

                SetRigidOriginal("LowerArmFront", "inertia");
                SetRigidOriginal("UpperArmFront", "inertia");
                return;

            }

            LB["UpperArmFront"].Broken   = true;

            if (LB["Head"].SpeciesIdentity == "Android")
            {

            }
            else
            {
                RB2["UpperArmFront"].inertia =
                RB2["LowerArmFront"].inertia = 0.00125f;

            }

            Vector3 diff = ((Vector3)RB2["LowerArmFront"].position - Camera.main.ScreenToWorldPoint(Input.mousePosition));
            
            //  Prevent spasticated behavior
            float sqrmag = diff.sqrMagnitude;

            if( sqrmag < 105f ) { 
                if (pauseAiming) return;
                if( diff.sqrMagnitude < 101f ) {
                    pauseAiming = true;
                    KB.EnableMouse();
                    return;
                }
            }

            if (pauseAiming) KB.DisableMouse();
            pauseAiming = false;
            
            Vector3 angleVelocity;
            if (FacingLeft) angleVelocity = new Vector3(0f, 0f, (Mathf.Atan2(diff.y, diff.x) * Mathf.Rad2Deg) - (85.5f + (16.5f * Facing)));
            else angleVelocity            = new Vector3(0f, 0f, (Mathf.Atan2(diff.y, diff.x) * Mathf.Rad2Deg) - (85.5f + (10.5f * Facing)));
            RB2["LowerArmFront"].MoveRotation(Quaternion.Euler(angleVelocity));

        }

        //
        // ─── POINT WEAPON ────────────────────────────────────────────────────────────────
        //
        public void PointWeapon()
        {
            if (IsInVehicle)
            {
                if (HoldingF == null)
                {
                    IsPointing = false;
                    CheckMouseClickStatus();
                }
                if (HoldingF?.JointType == Thing.JointTypes.FixedJoint) HoldingF?.ChangeJoint(Thing.JointTypes.HingeJoint);
                
                Vector3 mouseX = Camera.main.ScreenToWorldPoint(Input.mousePosition);
                Vector2 TPos;
                float TAngle;

                if (mouseX.x * Facing > ((RB2["Head"].position.x + 1.5) * Facing))
                {
                    TPos   = (RB2["UpperBody"].position + (Vector2.right * 3) - HoldingF.R.position);
                    TAngle = Vector2.SignedAngle(Vector2.right, TPos) - (HoldingF.angleAim + (HoldingF.angleOffset * Facing));
                    
                } else
                {
                    TPos   = (mouseX - HoldingF.tr.position);
                    TAngle = Vector2.SignedAngle(Vector2.right, TPos) - (HoldingF.angleAim + (HoldingF.angleOffset * Facing));
                }
                
                HoldingF?.R.MoveRotation(TAngle);
                HoldingF?.R.AddForce(TPos * TotalWeight * Time.deltaTime * 12f);

                if (KB.Throw)
                {
                    Actions.throwItem.power = Garage.Bike.CurrentSpeed;
                    Actions.throwItem.Init();
                }

                return;
            }
            else
            if (HoldingF == null || HoldingF?.isChainSaw == false)
            {
                IsPointing                    =
                LB["UpperArm"].Broken      =
                LB["UpperArmFront"].Broken =
                LB["LowerArm"].Broken      =
                LB["LowerArmFront"].Broken = false;
                RunRigids(RigidReset);
                return;
            }

            Vector3 mouse = Camera.main.ScreenToWorldPoint(Input.mousePosition);
            if (mouse.x * Facing > (RB2["Head"].position.x * Facing))
            {
                LB["UpperArm"].Broken       =
                LB["UpperArmFront"].Broken  =
                LB["LowerArm"].Broken       =
                LB["LowerArmFront"].Broken  = false;
                RunRigids(RigidReset);
                return;
            }

            LB["UpperArm"].Broken           =
            LB["UpperArmFront"].Broken      =
            LB["LowerArm"].Broken           =
            LB["LowerArmFront"].Broken      = (IsWalking || IsCrouching);

            HoldingF.R.inertia                 = 0.01f;
            HoldingF.R.mass                    = HoldingF.P.InitialMass / 2;

            if (pauseAiming) return;

            Vector3 TargetPos = (mouse - HoldingF.tr.position);

            float TargetAngle = Vector2.SignedAngle(Vector2.right, TargetPos) - (HoldingF.angleAim + (HoldingF.angleOffset * Facing));

            HoldingF.R.AddForce(TargetPos * TotalWeight * 3.5f );

            RB2["UpperBody"].AddForce(TargetPos * TotalWeight * 3.5f * -1);
        }



        //
        // ─── INVINCIBLE ────────────────────────────────────────────────────────────────
        //
        public void Invincible(bool doit)
        {
            foreach (LimbBehaviour limb in LB.Values) { if (limb.isActiveAndEnabled) limb.ImmuneToDamage = doit; }
        }

        //
        // ─── RUN LIMBS ────────────────────────────────────────────────────────────────
        //
        public void RunLimbs(Action<LimbBehaviour> action)
        {
            foreach (LimbBehaviour limb in LB.Values) { if (limb.isActiveAndEnabled) action(limb); }
        }

        public void RunLimbs<t>(Action<LimbBehaviour, t> action, t option)
        {
            foreach (LimbBehaviour limb in LB.Values) { if (limb.isActiveAndEnabled) action(limb, option); }
        }

        public void LimbFireProof(LimbBehaviour limb, bool option) {  
            
            PhysicalProperties propys           = limb.GetComponent<PhysicalBehaviour>().Properties.ShallowClone();
            propys.Flammability                 = option ? 0f : LimbOriginals[limb.name].Properties.Flammability;
            propys.BurningTemperatureThreshold  = float.MaxValue;
            propys.Burnrate                     = 0.00001f;

            limb.PhysicalBehaviour.Extinguish();
            limb.PhysicalBehaviour.Properties          = propys;
            limb.PhysicalBehaviour.SimulateTemperature = !option;
            limb.PhysicalBehaviour.ChargeBurns         = false;
            limb.PhysicalBehaviour.BurnProgress        = 0f;
            limb.DiscomfortingHeatTemperature          = float.MaxValue;
        }
        
        public void LimbImmune(LimbBehaviour limb, bool option)     => limb.ImmuneToDamage = option;
        public void LimbHeal(LimbBehaviour limb) { limb.BruiseCount = 0; }
        public void LimbGhost (LimbBehaviour limb, bool option) { limb.gameObject.layer = LayerMask.NameToLayer(option ? "Debris" : "Objects"); }
        public void LimbStrenthen (LimbBehaviour limb) { 
            if (limb.BaseStrength < 15f) { 
                limb.BaseStrength = Mathf.Min(15f, limb.BaseStrength + 5f); 
                limb.Wince(10f);
            }
        }

        public void LimbRecover(LimbBehaviour limb)
        {
            limb.HealBone();
            limb.CirculationBehaviour.HealBleeding();

            limb.Health                                 = limb.InitialHealth;
            
            limb.Numbness                               =
            limb.PhysicalBehaviour.BurnProgress         =
            limb.SkinMaterialHandler.AcidProgress       =
            limb.SkinMaterialHandler.RottenProgress     = 0.0f;
            
            limb.CirculationBehaviour.IsPump            = limb.CirculationBehaviour.WasInitiallyPumping;
            limb.CirculationBehaviour.BloodFlow         = 1f;
            
            limb.BruiseCount                            = 
            limb.CirculationBehaviour.StabWoundCount    =
            limb.CirculationBehaviour.GunshotWoundCount = 0;

            limb.CirculationBehaviour.ForceSetAllLiquid(0f);
            limb.CirculationBehaviour.AddLiquid(limb.GetOriginalBloodType(), 1f);
            limb.CirculationBehaviour.AddLiquid(LifeLiquid, 0.1f);
        }

        public void LimbGangster (LimbBehaviour limb)
        {

            limb.RegenerationSpeed = 
            limb.BreakingThreshold = float.MaxValue;
            
            limb.LungsPunctured = 
            limb.IsLethalToBreak = false;
            
            limb.CirculationBehaviour.InternalBleedingIntensity =
            limb.Vitality = 
            limb.ImpactPainMultiplier = 
            limb.ShotDamageMultiplier = 
            limb.Numbness *= 0f;
            
            limb.BruiseCount = 0;

            if (limb.Health < limb.InitialHealth) limb.Health = limb.InitialHealth;
        }

        public void LimbReset(LimbBehaviour limb)
        {
            LimbSnapshot OG = LimbOriginals[limb.name];

            limb.BaseStrength      = OG.BaseStrength;
            limb.Broken            = OG.Broken;
            limb.BreakingThreshold = OG.BreakingThreshold;
            limb.BruiseCount       = OG.BruiseCount;
            limb.Frozen            = OG.Frozen;
            limb.Health            = OG.Health;
            limb.Numbness          = OG.Numbness;
            limb.RegenerationSpeed = OG.RegenerationSpeed;
            limb.Vitality          = OG.Vitality;

            //BaseStrength = limb.BaseStrength,
            //        BreakingThreshold = limb.BreakingThreshold,
            //        Broken = limb.Broken,
            //        BruiseCount = limb.BruiseCount,
            //        Frozen = limb.Frozen,
            //        Health = limb.Health,
            //        Numbness = limb.Numbness,
            //        RegenerationSpeed = limb.RegenerationSpeed,
            //        Vitality = limb.Vitality,
        }




        //
        // ─── RUN RIGIDS ────────────────────────────────────────────────────────────────
        //
        public void RunRigids(Action<Rigidbody2D> action)
        {
            foreach (Rigidbody2D rigid in RB2.Values) { action(rigid); }
        }

        public void RunRigids<t>(Action<Rigidbody2D, t> action, t option)
        {
            foreach (Rigidbody2D rigid in RB2.Values) { action(rigid, option); }
        }

        public void RigidInertia(Rigidbody2D rb, float option) => rb.inertia = (option == -1) ? RigidOriginals[rb.name].inertia : option;
        public void RigidMass(Rigidbody2D rb, float option) => rb.mass = (option == -1) ? RigidOriginals[rb.name].mass : option;
        public void RigidAddMass(Rigidbody2D rb, float option) => rb.mass *= option;
        public void RigidDrag(Rigidbody2D rb, float option) => rb.drag = (option == -1) ? RigidOriginals[rb.name].drag : option;
        public void BodyInertiaFix(Rigidbody2D rb) => SetRigidOriginal(rb.name);
        public void RigidReset(Rigidbody2D rb)
        {
            rb.mass    = RigidOriginals[rb.name].mass;
            rb.drag    = RigidOriginals[rb.name].drag;
            rb.inertia = RigidOriginals[rb.name].inertia;
        }
        public void RigidStop(Rigidbody2D rb)
        {
            rb.velocity        = Vector2.zero;
            rb.angularVelocity = 0f;
        }


        

        //
        // ─── HOLD THING ────────────────────────────────────────────────────────────────
        //
        public void HoldThing(Thing thing)
        {
            if (HoldingF != null)
            {
                DropThing();
                return;
            }

            HoldingF = thing;
            HoldingF.AttachPuppetHand(this, true);

            if (Actions.CombatMode && HoldingF.canShoot) Actions.CombatMode = false;
            if (IsAiming && !HoldingF.canShoot) AimingStop();

            if (IsInVehicle) Garage.NoCollisionsWithHeldItem(HoldingF);

            foreach (Transform tr in customParts.Transforms)
            {
                if (tr != null) Util.ToggleCollisions(HoldingF.tr, tr, false, false);
            }
        }

        public void DropThing()
        {
            if (HoldingF != null)
            {
                if (HoldingF.P == null)
                {
                    HoldingF = (Thing)null;
                    return;
                }

                if (GripF.isHolding)
                {
                    LB["LowerArmFront"].SendMessage("Use", (object)new ActivationPropagation(), SendMessageOptions.DontRequireReceiver);
                    if (LB["LowerArm"].GripBehaviour.isHolding)
                    {
                        LB["LowerArm"].SendMessage("Use", (object)new ActivationPropagation(), SendMessageOptions.DontRequireReceiver);
                    }
                    Util.DisableCollision(PBO.transform, HoldingF.P, true);

                    PuppetMaster.CheckDisabledCollisions = true;

                    HoldingF.Dropped();
                    HoldingF = (Thing)null;
                }
            }
        }

        //
        // ─── RECOVER ────────────────────────────────────────────────────────────────
        //
        public void Recover()
        {
            PBO.Consciousness   = 1f;
            PBO.ShockLevel      = 0.0f;
            PBO.PainLevel       = 0.0f;
            PBO.OxygenLevel     = 1f;
            PBO.AdrenalineLevel = 1f;

            if (KB.Modifier) RunLimbs(LimbRecover);
        }


        //
        // ─── SET RIGID ORIGINAL ────────────────────────────────────────────────────────────────
        //
        public void SetRigidOriginal(string rigidName, string propName="")
        {
            List<string> props = new List<string>() { "mass", "drag", "inertia" };

            if (propName != "") {
                props.Clear();
                props.Add(propName);
            }

            foreach (string pname in props)
            {
                if (RigidOriginals.TryGetValue(rigidName, out RigidSnapshot rigidOG))
                {
                    switch(pname)
                    {
                        case "mass":
                            RB2[ rigidName ].mass = rigidOG.mass;
                            break;

                        case "drag":
                            RB2[rigidName].drag = rigidOG.drag;
                            break;

                        case "inertia":
                            RB2[rigidName].inertia = rigidOG.inertia;
                            break;
                    }

                }
            }
        }

        public void CheckMouseClickStatus()
        {
            if (!IsAiming && !Actions.CombatMode && !IsHeroAiming && !IsPointing) KB.EnableMouse();
        }


        
    }
}
